#include "oms_types.h"
#include "debug.h"
#include "basis_equation.h"
#include "globaldef.h"
#include "nodedef.h"
#include "combinat.h"

static int PL_STATE_MASK[]={2,0,1};
typedef int small_int;

#include "pluecker_state.h"


static void pluecker_partitions(be_block_p,array_index_t, array_index_t, array_index_t,
				array_index_t, natural_t *, natural_t *, 
				array_index_t, array_index_t, array_index_t);

static void find_pluecker_bases(be_block_p,size_t,size_t);


/** 
 * Initialize an appropriately sized basis equation block for the 
 * Pl\"ucker equations.
 */
extern void pl_init(be_block_p PE, natural_t n, natural_t r){
  natural_t num_bases,num_eqns;

  num_bases=binomial(n,r);
  num_eqns=binomial(n,r+2)*binomial(r+2,4); 

  err_assert(PE->cols==6);
  err_assert(PE->bases==num_bases);
  err_assert(PE->rows==num_eqns);
  err_assert(PE->eqn !=NULL);

  PE->stride=2;

  find_pluecker_bases(PE,n,r);

  be_makelinks(PE);  
}


/** 
 * Compute the basis index representation of the Pl\"ucker equations.
 * These are also known as the Chirotope axioms.
 * @param n number of points.
 * @param d rank of the oriented matroid. 
 */

static void find_pluecker_bases(be_block_p PE,size_t n, size_t d){
  natural_t *points;		/* set of @d+2@ point indices */
  array_index_t quad[4];		/* set of 4 indices into @points@ */
  size_t j;			/* Count of equations found so far */
  int sigma;

  points=calloc(d+4,sizeof(natural_t));
  ERR_TRAP_NULL(points);

  j=0;
  first_kset(n,d+2,points);	/* @points =< 1..d+2 >@ */
  do {
    first_kset(d+2,4,quad);	/* @quad = < 1...4 > @ */
    do {
      /*%
       * For each $d+2$-set $P$, for each subset $Q\subset P$
       * of 4 vertices, we partition all possible ways such that $2$
       * vertices of $Q$ stay and two leave.
       %*/
      pluecker_partitions(PE,n,d,    j,0,    points,quad,1,2,3);
      pluecker_partitions(PE,n,d,    j,2,    points,quad,2,1,3);
        /* the inner term is negated; we handle this by negating 
	   one of the literals */
      sigma=be_sign(PE,j,2);
      be_set_sign(PE,j,2,-1*sigma);

      pluecker_partitions(PE,n,d,    j,4,    points,quad,3,1,2);
      j++;
    } while (next_kset(d+2,4,quad));
  } while (next_kset(n,d+2,points));
  free(points);
}


/** 
 * Traverse pluecker equations

 * 
 * This is the more optimized of the two traversal routines, but
 * assume that every sign chosen is run through the pluecker
 * equations. In particular it stores the encoded signs of each
 * equation in pluecker_state, even though those are implicit in the
 * chirotope

 * @param k which basis to check
 * @param BE Basis equations
 * @param Q (output) forced basis signs
 * @param chi chirotope
 * @param pluecker_state state of basis equations (what is set)
 * 
 * @return true if ok
 */
bool pl_traverse(array_index_t k,
			  be_block_t *BE,signed_set_p S,
			  sarray_p chi,sa16_array_p pluecker_state) {
  int elem;


  elem=BE->last[k];		/* @elem@ points to the last occur. of @k@ */
  err_assert( elem >= -1 && elem < BE->rows*BE->cols );
 


  while ( elem>=0  ){
    array_index_t row,col,fcol,row_start,new_sign;
    uint16_t row_state;
    be_coeff_p Eqn,entry;

    entry=BE->eqn+elem;

    row=be_row(BE,elem);
    row_start=be_row_start(BE,elem);
    col=be_col(BE,elem);

    Eqn=BE->eqn+row_start;	/* 1.8% */

    new_sign=sa_read(chi,be_bindex_raw(BE,elem))*
		     be_sign_raw(BE,elem); /* 2.53% */

    row_state=sa16_or2(pluecker_state,row,	
			(PL_STATE_MASK[new_sign+1] << col*2)
			);


    fcol=forced_col[row_state];	/* 2.16% */
    err_assert( fcol >= -1  && fcol < 6);

    if (fcol != -1){
      /* There is exactly one nonzero column */
      int fsign=be_sign_raw(BE,row_start+fcol)*forced_sign[row_state];
      if (fsign != 0) {
	int fbasis=be_bindex_raw(BE,row_start+fcol);
	int cur_sign=sa_read(chi,fbasis);
	
	if (cur_sign !=0 && cur_sign !=fsign)
	  return FALSE;
	if (cur_sign==0){
	  err_debug(DB_PLUECKER,(HERE,"forcing %d to %d",fbasis,fsign));
	  if (!ss_add(S,fbasis,fsign))
	    return FALSE;
	}

      } else {
	/* We want to force something, but we can't */
	if (fsign != 0)
	  return FALSE;
      }
    }

    elem=BE->eqn[elem].prev;	/* 1.8 % */
  }
  return TRUE;
}

/**
 * Traverse pluecker equations.


   This routine is not as fast as pl_traverse, but
   it does not rely on a stored pluecker_state
*/
bool pl_traverse2(array_index_t bindex, na_node_p local,
			   na_node_p global,signed_set_p S){

  int elem;
  be_block_p PE;

  sarray_p chi;

  err_debug(DB_PLUECKER,("pl_traverse2","%d",bindex));

  PE=na_access(global,SLOT_pe);
  chi=na_access(local,SLOT_chirotope);

  err_assert(bindex < PE->bases);

  elem=PE->last[bindex];     /* @elem@ points to the last occur. of @bindex@ */

  err_assert( elem >= -1 && elem < PE->rows*PE->cols );

  while ( elem>=0  ){
    array_index_t fcol,row_start,new_sign;
    int row_state,j;
    be_coeff_p Eqn,cell;
    
    cell=PE->eqn+elem;

    row_start=be_row_start(PE,elem);

    Eqn=PE->eqn+row_start;	

    row_state=0;

    for (j=0; j<6; j++){
      new_sign=sa_read(chi,be_bindex_raw(PE,row_start+j))*
		       be_sign_raw(PE,row_start+j);
      row_state |= (PL_STATE_MASK[new_sign+1] << j*2); 
    }

    fcol=forced_col[row_state];	/* 2.16% */
    err_assert( fcol >= -1  && fcol < 6);

    if (fcol != -1){
      /* There is exactly one nonzero column */
      int fsign=be_sign_raw(PE,row_start+fcol)*forced_sign[row_state];
      if (fsign != 0) {
	int fbasis=be_bindex_raw(PE,row_start+fcol);
	int cur_sign=sa_read(chi,fbasis);
	
	if (cur_sign !=0 && cur_sign !=fsign)
	  ERR_RETURN(FALSE);
	if (cur_sign==0){
	  err_debug(DB_PLUECKER,(HERE,"forcing %d to %d",fbasis,fsign));
	  ss_add(S,fbasis,fsign);
	}

      } else {
	/* We want to force something, but we can't */
	if (fsign != 0)
	  ERR_RETURN(FALSE);
      }
    }

    elem=PE->eqn[elem].prev;	
  }
  ERR_RETURN( TRUE);
}

/* pi should have two extra spaces at the end, initially zero */
static void pluecker_partitions(be_block_p PE,
				array_index_t n, array_index_t dim,
				array_index_t row,
			array_index_t i,
			natural_t *pi,  natural_t *q,
			array_index_t b, array_index_t c, array_index_t d){

  natural_t t1,t2;
  int sigma;

  err_assert(row >=0);
  err_assert(i >= 0 && i <= 4);
  
  t1=0; t2=0;

  /* keep a, b, delete c,d */

  swap(t1,pi[q[c]-1]);  swap(t2,pi[q[d]-1]);
  swap(pi[q[0]-1],pi[dim+2]);   swap(pi[q[b]-1],pi[dim+3]);
  be_set_bindex(PE,row,i,kset_cindex(n,dim,pi));
  be_set_sign(PE,row,i,basis_parity(dim,pi));
  swap(t1,pi[q[c]-1]);  swap(t2,pi[q[d]-1]);
  swap(pi[q[0]-1],pi[dim+2]);   swap(pi[q[b]-1],pi[dim+3]);

  /* Now the other way around */
  swap(pi[q[c]-1],pi[dim+2]);   swap(pi[q[d]-1],pi[dim+3]);
  swap(t1,pi[q[0]-1]);  swap(t2,pi[q[b]-1]);  

  be_set_bindex(PE,row,i+1,kset_cindex(n,dim,pi));
  be_set_sign(PE,row,i+1,basis_parity(dim,pi));

  swap(t1,pi[q[0]-1]);  swap(t2,pi[q[b]-1]);  
  swap(pi[q[c]-1],pi[dim+2]);   swap(pi[q[d]-1],pi[dim+3]);
  

}




