/**
 * @file   runtime_options.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Wed Jan 22 21:18:44 2003
 * 
 * @brief  process options that can be both command line and in input
 * 
 * 
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


extern FILE *yyin;

#include "output.h"
#include "error_message.h"
#include "globaldef.h"
#include "gparams.h"
#include "nalloc.h"
#include "nu_sign.h"
#include "runtime_options.h"
#include "state.tab.h"
#include "startopt.h"
#include "oms_opt.h"
  
static params_p params=0;


static bool ro_process_zram_arg(int optchar, params_p params);
static void ro_usage(char **argv, startopt_t startopt);


#include "option_funcs.c"

void ro_init(params_p global_params,startopt_t startopt){

  /* save for access from parser */

  ro_set_param_block(global_params);

  ro_set_defaults(params,startopt);


  /* parameters not defined in runopts.def */

  params->subtreelimit=OMS_SUBTREELIMIT;
  params->checkpointtime=OMS_CHECKPOINTTIME;
  params->checkpointname[0]=0;
  params->restart=0;
  params->variable_order='l';
  params->inputfile[0]=0;


}

extern void ro_process_args(int argc, char **argv, 
			    params_p params,startopt_t startopts){
  int c,errflag=0;
  bool ok;

  optarg=NULL;

  ro_init(params,startopts);

  strncpy(params->myname,argv[0],PATH_MAX);

  while (!errflag && ((c = getopt(argc, argv, RO_GETOPT_STR ))
		      != -1)){
    switch (c) {
    case 'i':
      strncpy(params->inputfile,optarg,PATH_MAX);
      yyin=fopen(optarg,"r");
      err_assert(yyin);
      break;
    case 'o':
      oms_redir_output(optarg);
      break;
    case 'r':
      params->restart=1;
      strncpy(params->restartname,optarg,PATH_MAX);
      break;
    
    case 'V':
      params->variable_order=optarg[0];
      break;
    case 'h':
      ro_usage(argv,startopts);
      exit(1);
    default: 
      ok=ro_process_auto_arg(c,optarg);
      if (!ok && OPT_use_zram)
	ok = ok || ro_process_zram_arg(c,params);
      
      if (!ok)	{
	ro_usage(argv,startopts);
	exit(1);
      }
    }
  }
}

static bool ro_process_zram_arg(int optchar,params_p params){
  switch(optchar){
  case 'c':
    strncpy(params->checkpointname,optarg,PATH_MAX);
    break;
  case 'l':
    params->subtreelimit=atoi(optarg);
    break;
  case 't':
    params->checkpointtime=atol(optarg);
    break;
  default:
    return FALSE;
  }
  return TRUE;
}

void ro_usage_auto(){
  int i;

  for (i=0; usage_msg[i]; i++){
    fputs(usage_msg[i],stderr);
    fputs("\n",stderr);
  }
}
void ro_set_param_block(params_p global_params){
  params=global_params;
}


static void ro_usage(char **argv, startopt_t startopt){
  fprintf(stderr,"usage: %s: [options]\n",argv[0]);
  
  fprintf(stderr,"\nInput/Output:\n");
  fprintf(stderr,"-i <file>\ttake input from <file>\n");
  fprintf(stderr,"-o <file>\twrite input to <file>\n");
  
  ro_usage_auto();
  
  if (OPT_use_zram  && (startopt & SO_pluecker) ){
    fprintf(stderr,"\nLoad Balancing:\n");
    fprintf(stderr,"-l <n>\tLimit sequential subtree to n nodes (default=%d)\n",OMS_SUBTREELIMIT);
    fprintf(stderr,"-D <n>\tLimit sequential subtree to depth n (default no limit)\n");
    fprintf(stderr,"-1\t run sequentially\n");
    fprintf(stderr,"\nCheckpointing\n");
    fprintf(stderr,"-c <name>\tUse <name> for checkpoint files (default checkXXXXXX)\n");
    fprintf(stderr,"-r <name>\trestart from <name>\n");
    fprintf(stderr,"-t <n>\tCheckpoint ever <n> seconds (default %d)\n",
	    OMS_CHECKPOINTTIME);
  }
  
  if (startopt & SO_pluecker){
    fprintf(stderr,"\nSearch Heuristics\n");
    fprintf(stderr,"-V <c>\t Use variable ordering strategy <c> (default l)\n");
    fprintf(stderr,"\tg:\tgreedy\n");
    fprintf(stderr,"\tl:\tlexicographic\n");  
    fprintf(stderr,"\tL:\treverse lexicographic\n");  
    fprintf(stderr,"\tr:\trandom\n");
  } 

  if (startopt & SO_hyperline){
    fprintf(stderr,"\n\t-x\textra pluecker checks (for hyperline search)\n");
  }

  fputs("\n",stderr);

}
