/* z_bb1.c (ZRAM: Branch and Bound in parallel)
   Ambros Marzetta

 * $Source$
 * $Revision$
 * $Date$
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.9  1996/04/09  13:34:31  marzetta
 * removed z_stdio.h
 *
 * Revision 1.8  1996/04/09  13:16:26  marzetta
 * new InstallJob
 *
 * Revision 1.7  1995/08/22  12:39:12  marzetta
 * New message passing, MPI
 *
 * Revision 1.6  1995/08/02  13:28:46  marzetta
 * Portability to macintosh
 *
 * Revision 1.5  1995/07/31  14:43:48  marzetta
 * Branch in place
 *
 * Revision 1.4  1995/07/28  11:46:47  bruengg
 * fixed node count with initial upper bound
 *
 * Revision 1.3  1995/07/27  11:36:00  marzetta
 * Separation of message passing interface
 *
 */

#include <stdio.h>
#include <stdlib.h>		/* exit */
#include "z_base.h"
#include "z_bb.h"
#include "z_spq.h"
#include "z_par.h"
#include "z_mp.h"
#include "z_time.h"		/* front end only */

/*#define TRACENODE 10000*/
#define TRACENEWUB0

static boolean bbp_inited;

#ifndef SEQUENTIAL

#define ShortNodeString(x) glob->psp->bb_ShortNodeString ? \
  glob->psp->bb_ShortNodeString(x) : ""
  
#define ShortNodeString2(x) glob.psp->bb_ShortNodeString ? \
  glob.psp->bb_ShortNodeString(x) : ""
  
/* Branch and bound parallel -------------------------------------------- */

/* Issues: only send MINTs between FE and PE. */

#define BALANCE_FREQ 40

static z_JobNum z_bb_parallelJob;
static z_tagT newUBTag;

typedef struct PglobalsT {
  z_bb_ProcSetP psp;
  z_NodeP solNode;		/* upper bound */
  z_bb_GlobalDataP globalData;
  z_spqP spq;			/* speculative priority queue */
  int queueSize;
  long branchCount;
  #ifdef TRACENODE
    long lastBranchCountD, lastBranchCountB;
  #endif
  boolean noisy;
  int tag;
  long numNodesStart;		/* nodes at start + solNode */
  boolean ub_changed;		/* 1 when new ub found; 0 when broadcast */
				/* always 0 if there is only one PE */
  
  int startup;			/* > 0 during startup phase */
  long next_balance;		/* when to send the next nodes away */
  int next_balance_dir;		/* in which direction to send it */
  int has[MAX_PE];		/* how many nodes the other PEs have */
  int numhas;			/* number of entries != -1 in has[] */
} PglobalsT;

static PglobalsT glob;

static void NewUBHandler(void)
     /* message handler */
{
  z_NodeP newub;

#ifdef TRACENEWUB
    printf("[%2d]<New Upper Bound\r\n", z_myproc);
#endif
    newub = z_GetNodeBuffer();
    if (glob.solNode == NULL ||
	glob.psp->bb_CompareProc(newub, glob.solNode) < 0) {
      /* received node is better than local upper bound */
      if (glob.solNode)
	z_FreeNode(glob.solNode);
      else
	glob.numNodesStart ++;
      glob.solNode = newub;
      z_spq_DeleteLocalAbove(glob.spq, newub,
			     (whatToDoProcP) z_FreeNode);
    } else
      z_FreeNode(newub);
} /* NewUBHandler */

static void BroadcastUB(void)
{
  /* Broadcast the upper bound.
     Precondition: z_TestSend
     One node should always fit into the output buffer.
     */
  /*printf("Broadcasting upper bound\n");*/
  z_AddTagBuffer(newUBTag);
  z_AddNodeBuffer(glob.solNode);
  z_Ibroadcast("upper bound");
  glob.ub_changed = 0;
} /* BroadcastUB */

static void DoPChildInPlace(z_NodeP child, PglobalsT *glob);

static void DoPChild(z_NodeP child, PglobalsT *glob)
{
  /* This procedure is called for all children. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */
  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
    /* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode)
	z_FreeNode(glob->solNode);
      else
	glob->numNodesStart ++;
      glob->solNode = child;
      if (!glob->ub_changed && z_numprocs > 1) z_QueueSender(BroadcastUB);
      glob->ub_changed = z_numprocs > 1;
      /*if (z_TestSend()) BroadcastUB();*/
      if (glob->noisy) {
	printf("[%2d]New upper bound: (%s)\r\n", z_myproc,
	       ShortNodeString(child));
	fflush(stdout);
      }
      z_spq_DeleteLocalAbove(glob->spq, child,
			    (whatToDoProcP) z_FreeNode);
    } else { /* better than upper bound, but no solution */
      if (z_spq_numLocalElements(glob->spq) < glob->queueSize) { /* insert */
	z_spq_Insert(glob->spq, child); /* no solution -> queue */
	/*printf("    Insert (%s)\r\n",
	  ShortNodeString(child));*/
      } else { /* continue depth first */
	z_Poll();
        #ifdef TRACENODE
	  if (glob->branchCount - glob->lastBranchCountD >= TRACENODE) {
	    glob->lastBranchCountD = glob->branchCount;
	    printf("[%2d]  Node (depth first): (%s)\r\n", z_myproc,
		   ShortNodeString(child));
	    fflush(stdout);
	  }
        #endif /* TRACENODE */
	glob->branchCount++;
	if (glob->psp->bb_BranchInPlace)
	  glob->psp->bb_BranchInPlace(glob->globalData,
				      child, glob->solNode,
				      (z_bb_DoProcP) DoPChildInPlace, glob);
	else
	  glob->psp->bb_Branch(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoPChild, glob);
	z_FreeNode(child);
      } /* continue depth first */
    } /* better than upper bound, but no solution */
  } /* better than upper bound */
  else {
    z_FreeNode(child); /* not better than upper bound -> Free */
  }
} /* DoPChild */

static void DoPChildInPlace(z_NodeP child, PglobalsT *glob)
{
  /* This procedure is called for all children. The child either goes to
     glob->solNode, to the priority queue, or is deallocated.
     */

  if (glob->solNode == NULL ||
      glob->psp->bb_CompareProc(child, glob->solNode) < 0) {
    /* better than upper bound */
    if (glob->psp->bb_IsSolution(child)) { /* solution -> new upper bound*/
      if (glob->solNode)
	z_FreeNode(glob->solNode);
      else
	glob->numNodesStart ++;
      glob->solNode = z_NewCopyNode(child);
      if (!glob->ub_changed && z_numprocs > 1) z_QueueSender(BroadcastUB);
      glob->ub_changed = z_numprocs > 1;
      /*if (z_TestSend()) BroadcastUB(glob);*/
      if (glob->noisy) {
	printf("[%2d]New upper bound: (%s)\r\n", z_myproc,
	       ShortNodeString(child));
	fflush(stdout);
      }
      z_spq_DeleteLocalAbove(glob->spq, child,
			    (whatToDoProcP) z_FreeNode);
    } else { /* better than upper bound, but no solution */
      if (z_spq_numLocalElements(glob->spq) < glob->queueSize) { /* insert */
	z_spq_Insert(glob->spq, z_NewCopyNode(child));
				/* no solution -> queue */
      } else { /* continue depth first */
	z_Poll();
        #ifdef TRACENODE
	  if (glob->branchCount - glob->lastBranchCountD >= TRACENODE) {
	    glob->lastBranchCountD = glob->branchCount;
	    printf("[%2d]  Node (depth first): (%s)\r\n", z_myproc,
		   ShortNodeString(child));
	    fflush(stdout);
	  }
        #endif /* TRACENODE */
	glob->branchCount++;
	if (glob->psp->bb_BranchInPlace)
	  glob->psp->bb_BranchInPlace(glob->globalData,
				      child, glob->solNode,
				      (z_bb_DoProcP) DoPChildInPlace, glob);
	else
	  glob->psp->bb_Branch(glob->globalData,
			       child, glob->solNode,
			       (z_bb_DoProcP) DoPChild, glob);
      } /* continue depth first */
    } /* better than upper bound, but no solution */
  } /* better than upper bound */
} /* DoPChildInPlace */

static void z_BBP(PglobalsT *glob)
{
  /* If there is a better solution in the given subtree, adjust
     solNode. Deallocate the old contents of solNode.
     Allocate the new contents.
     */

  z_NodeP node;
  boolean terminate;

  terminate = 0;
  while (! terminate) {
    /* There is a receive pending */
    z_Poll();
    /* There is a receive pending */
    assert(glob->numNodesStart ==
	   z_numNodes[glob->tag] - z_spq_numLocalElements(glob->spq));
    
    node = z_spq_GetDeleteNearMinimum(glob->spq);
    
    terminate = ! node;
    
    if (! terminate) {
      /* node may be higher than upper bound if it has crossed the
	 newub message ????????????? */
      
      assert(! glob->psp->bb_IsSolution(node));
      /* not a leaf -> branch */
      glob->branchCount++;
#ifdef TRACENODE
      if (glob->branchCount - glob->lastBranchCountB >= TRACENODE) {
	glob->lastBranchCountB = glob->branchCount;
	if (glob->noisy)
	  printf("[%2d]  Node (queue=%5ld): (%s)\r\n",
		 z_myproc, z_spq_numLocalElements(glob->spq),
		 ShortNodeString(node));
	fflush(stdout);
      }
#endif /* TRACENODE */
      if (glob->psp->bb_Branch)
	glob->psp->bb_Branch(glob->globalData, node, glob->solNode,
			     (z_bb_DoProcP) DoPChild, glob);
      else
	glob->psp->bb_BranchInPlace(glob->globalData, node, glob->solNode,
				    (z_bb_DoProcP) DoPChildInPlace, glob);
      z_FreeNode(node);
    } /* if (! terminate) */
  } /* while (! terminate) */
  /* no receive pending */
} /* z_BBP */

static void z_bb_parallel(void)
{
  /* Parallel branch and bound: part running on processing elements. */
  MINT psn;
  z_NodeP root;

  {
    /* Read the message broadcast by the front end in z_BranchBoundParallel.
       It contains global data */
    glob.queueSize = z_GetMINTBuffer();
    psn = z_GetMINTBuffer();
    glob.noisy = z_GetMINTBuffer();
    glob.globalData = (z_bb_GlobalDataP) z_GetNodeBuffer();
    root = z_GetNodeBuffer();
    glob.solNode = z_GetNodeBuffer();
    glob.psp = (z_bb_ProcSetP) z_GetProcSet(psn);
  }
  
  z_StartMessagePassing();
  glob.tag = root->tag;
  glob.numNodesStart = z_numNodes[glob.tag];

  /* Compute the initial upper bound on the last processing element */
  glob.ub_changed = 0;
  if (!glob.solNode && (z_numprocs == 1 || z_myproc == z_numprocs - 1)) {
    if (glob.psp->bb_UpperBound)	/* could do this on front end */
      glob.solNode = glob.psp->bb_UpperBound(glob.globalData,root);
    if (glob.solNode) {
      glob.numNodesStart ++;
      glob.ub_changed = z_numprocs > 1;
      if (glob.ub_changed) z_QueueSender(BroadcastUB);
    }
  }
  if (glob.noisy && glob.solNode)
    printf("[%2d]Initial upper Bound (%s)\r\n", z_myproc,
	   ShortNodeString2(glob.solNode));
  
  glob.spq = z_spq_NewSpeculativePriorityQueue(
		  (z_pq_CompProcP)glob.psp->bb_CompareProc);
  if (z_myproc == ROOT_PE)
    (void) z_spq_Insert(glob.spq, z_NewCopyNode(root));
  glob.branchCount = 0;
  #ifdef TRACENODE
    glob.lastBranchCountD = 0;
    glob.lastBranchCountB = 0;
  #endif
  z_BBP(&glob);
  if (glob.noisy) {
    printf("[%2d]Branch: %ld\r\n", z_myproc, glob.branchCount);
  }
  z_spq_FreeSpeculativePriorityQueue(glob.spq);
  assert(glob.numNodesStart == z_numNodes[glob.tag]);
  z_FreeNode(root);
  z_FreeNode(glob.globalData);
  if (z_myproc == ROOT_PE) { /* send glob.solNode to front end; */
    z_AddNodeBuffer(glob.solNode);
  }
  if (glob.solNode) z_FreeNode(glob.solNode);
} /* z_bb_parallel */


#ifndef PEonly

z_NodeP z_BranchBoundParallel(z_ProcSetNum procsetnum,
			      z_bb_GlobalDataP global,
			      z_NodeP root, z_NodeP initialUpperBound,
			      int queueSize, boolean noisy)
{
  /* Parallel branch and bound: part running on the front end. */
  /* Pre: output buffer is empty */
  z_bb_ProcSetP psp;
  z_NodeP sol;

  assert(z_frontend);		/* can be called on the front end only */
  assert(bbp_inited);
  
  psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(psp->bb_Branch || psp->bb_BranchInPlace);
  assert(psp->bb_CompareProc);
  assert(psp->bb_IsSolution);
  if (psp->bb_IsSolution(root))
    return z_NewCopyNode(root);
  
  /* Broadcast: procsetnum, *global, *root */
  {
    MINT psn;

    psn = procsetnum;
    z_AddMINTBuffer(queueSize);
    z_AddMINTBuffer(psn);
    z_AddMINTBuffer(noisy);
    z_AddNodeBuffer(global);
    z_AddNodeBuffer(root);
    z_AddNodeBuffer(initialUpperBound);
    z_StartParallelJob(z_bb_parallelJob, NULL);
  }

  /* Wait / work */
  z_WaitWork();
  
  /* Get Results: *solution */
  sol = z_GetNodeBuffer();
  
  return sol;
} /* z_BranchBoundParallel */

#endif /* ! PEonly */
#endif /* SEQUENTIAL */

/* Initialisation ------------------------------------------------------- */

void z_bb1_Init(void)
{
  /* called on every processor by initproc() */

  if (! bbp_inited) {
    bbp_inited = 1;
    z_spq_init();
#ifndef SEQUENTIAL
    z_bb_parallelJob = z_InstallJob(z_bb_parallel);
    newUBTag = z_InstallMessage(NewUBHandler);
#endif /* SEQUENTIAL */
  }
} /* z_bb1_Init */
